#include "kernel.h"

// 接收三环传递来的通讯线程
HANDLE gThreadHandle = NULL;
bool gWorkThreadFlags = false;
HANDLE gEventHandle = NULL;
PKEVENT gKernelEvent = NULL;


VOID onUnload(_In_ struct _DRIVER_OBJECT* DriverObject) {
    UNREFERENCED_PARAMETER(DriverObject);
    KdPrint(("[yhy]: %s", __FUNCTION__));

    // 删除驱动符号链接
    UNICODE_STRING symLink;
    RtlInitUnicodeString(&symLink, SYMBOLIC_LINK);
    NTSTATUS status = IoDeleteSymbolicLink(&symLink);
    if (!NT_SUCCESS(status)) {
        KdPrint(("[yhy] [Error]: IoDeleteSymbolicLink Error! errorCode:%08x\n", status));
        return;
    }
    // 删除设备
    if (DriverObject->DeviceObject != NULL) {
        IoDeleteDevice(DriverObject->DeviceObject);
    }
    KdPrint(("[yhy]: onUnload Success!\n"));
    KdPrint(("[yhy]: Bye!\n"));
}

NTSTATUS DispatchCreate(_In_ struct _DEVICE_OBJECT* DeviceObject, _Inout_ struct _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    KdPrint(("[yhy]: %s\n", __FUNCTION__));
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    UNREFERENCED_PARAMETER(DeviceObject);
    return status;
}

NTSTATUS DispatchClose(_In_ struct _DEVICE_OBJECT* DeviceObject, _Inout_ struct _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    KdPrint(("[yhy]: %s\n", __FUNCTION__));
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    UNREFERENCED_PARAMETER(DeviceObject);
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    return status;
}

NTSTATUS DispatchRead(_In_ struct _DEVICE_OBJECT* DeviceObject, _Inout_ struct _IRP* Irp) {
    UNREFERENCED_PARAMETER(DeviceObject);
    NTSTATUS status = STATUS_SUCCESS;
    KdPrint(("[yhy]: %s\n", __FUNCTION__));
    // 需要把数据写入缓冲区
    PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation(Irp);
    // 需要缓冲区地址+请求读写的长度
    const char* text = "hello world!";
    //PVOID pBuffer = Irp->AssociatedIrp.SystemBuffer;
    PVOID pBuffer = MmGetSystemAddressForMdlSafe(Irp->MdlAddress, NormalPagePriority);

    ULONG nBufferSize = pIrpStack->Parameters.Read.Length;
    // 拷贝字符串
    ULONG len = min(nBufferSize, (ULONG)strlen(text));

    KdPrint(("[yhy]: DispatchRead() text: %s pBuffer:%p nBufferSize:%d len:%d", text, pBuffer, nBufferSize, len));

    RtlCopyMemory(pBuffer, text, len);

    // 发送一个测试
    KeSetEvent(gKernelEvent, IO_NO_INCREMENT, FALSE);

    // 完成请求
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = len;


    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;
}
NTSTATUS DispatchWrite(_In_ struct _DEVICE_OBJECT* DeviceObject, _Inout_ struct _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    KdPrint(("[yhy]: %s\n", __FUNCTION__));
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    UNREFERENCED_PARAMETER(DeviceObject);
    KdPrint(("[yhy]: DeviceObject->StackSize: %d\n", DeviceObject->StackSize));
    return status;
}

NTSTATUS handleOnCtlReadMemory(PIO_STACK_LOCATION pIrpStack, _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    PVOID inBuffer = Irp->AssociatedIrp.SystemBuffer;
    PVOID outBuffer = Irp->AssociatedIrp.SystemBuffer;
    // 先解
    ReadMemoryModel* readModel = (ReadMemoryModel*)inBuffer;
    ULONG bufferLen = max((ULONG)readModel->len, pIrpStack->Parameters.DeviceIoControl.OutputBufferLength);
    // 读取目标内存
    KdPrint(("[yhy]: DispatchControl inBuffer:%p targetAddress:%p\n", inBuffer, readModel->address));
    RtlCopyMemory(outBuffer, readModel->address, bufferLen);
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = bufferLen;
    return status;
}

NTSTATUS handleOnCtlWriteMemory(PIO_STACK_LOCATION pIrpStack, _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    PVOID inBuffer = Irp->AssociatedIrp.SystemBuffer;
    //PVOID outBuffer = Irp->AssociatedIrp.SystemBuffer;
    UNREFERENCED_PARAMETER(pIrpStack);
    WriteMemoryModel* writeModel = (WriteMemoryModel*)inBuffer;
    ULONG bufferLen = writeModel->len;
    RtlCopyMemory(writeModel->address, writeModel->buffer, bufferLen);
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = bufferLen;
    return status;
}

NTSTATUS handleOnCtlEnumProc(PIO_STACK_LOCATION pIrpStack, _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    // TODO...
    UNREFERENCED_PARAMETER(pIrpStack);
    UNREFERENCED_PARAMETER(Irp);
    return status;
}

/*
* 内核驱动创建一个命名的事件对象，用来主动向应用层进行通信
*/
NTSTATUS handleCreateNotificationEvent() {
    NTSTATUS status = STATUS_SUCCESS;
    UNICODE_STRING EventName = { 0 };
    RtlInitUnicodeString(&EventName, NOTIFICATION_OBJECT_NAME);
    gKernelEvent = IoCreateNotificationEvent(&EventName, &gEventHandle);
    KdPrint(("[yhy] createNotificationEvent! gKernelEvent: %p gEventHandle: %p\n", gKernelEvent, gEventHandle));
    if (gKernelEvent == NULL) {
        KdPrint(("[yhy] createNotificationEvent Fail!\n"));
        status = STATUS_UNSUCCESSFUL;
    }
    return status;
}

NTSTATUS handleCloseNotificationEvent() {
    KdPrint(("[yhy]: %s\n", __FUNCTION__));
    NTSTATUS status = STATUS_SUCCESS;
    // 关闭事件对象
    // 唤醒事件
    KeSetEvent(gKernelEvent, IO_NO_INCREMENT, FALSE);
    KdPrint(("[yhy] [Error]: NtClose! gEventHandle:%p\n", gEventHandle));
    status = NtClose(gEventHandle);
    if (!NT_SUCCESS(status)) {
        KdPrint(("[yhy] [Error]: NtClose Error! errorCode:%08x\n", status));
    }
    return status;
}

NTSTATUS DispatchControl(_In_ struct _DEVICE_OBJECT* DeviceObject, _Inout_ struct _IRP* Irp) {
    NTSTATUS status = STATUS_SUCCESS;
    UNREFERENCED_PARAMETER(DeviceObject);
    PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation(Irp);
    ULONG ControlCode = pIrpStack->Parameters.DeviceIoControl.IoControlCode;

    switch (ControlCode) {
    case IOCTL_READ_MEMORY: {
        status = handleOnCtlReadMemory(pIrpStack, Irp);
        break;
    }
    case IOCTL_WRITE_MEMORY: {
        status = handleOnCtlWriteMemory(pIrpStack, Irp);
        break;
    }
    case IOCTL_ENUMERATION_PROCESS: {
        status = handleOnCtlEnumProc(pIrpStack, Irp);
        break;
    }
    case IOCTL_KERNEL_CREATE_EVENT_OBJECT: {
        status = handleCreateNotificationEvent();
        break;
    }
    case IOCTL_CLOSE_EVENT_OBJECT: {
        status = handleCloseNotificationEvent();
        break;
    }
    }
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;
}

NTSTATUS DriverEntry(
    _In_ PDRIVER_OBJECT     DriverObject,
    _In_ PUNICODE_STRING    RegistryPath
)
{
    UNREFERENCED_PARAMETER(RegistryPath);
    // NTSTATUS variable to record success or failure
    NTSTATUS status = STATUS_SUCCESS;
    DriverObject->DriverUnload = onUnload;
    // 1. 绑定IRP 处理Dispatch例程
    DriverObject->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = DispatchClose;
    DriverObject->MajorFunction[IRP_MJ_READ] = DispatchRead;
    DriverObject->MajorFunction[IRP_MJ_WRITE] = DispatchWrite;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DispatchControl;

    // 2. 创建设备对象
    UNICODE_STRING deviceName;
    PDEVICE_OBJECT pDevice = NULL;
    RtlInitUnicodeString(&deviceName, KERNEL_REG_DRVICE_NAME);
    KdPrint(("[yhy] deviceName: %wZ", &deviceName));
    status = IoCreateDevice(DriverObject, 0, &deviceName, FILE_DEVICE_UNKNOWN, FILE_DEVICE_SECURE_OPEN, FALSE /*非独占*/,
        &pDevice);
    if (!NT_SUCCESS(status)) {
        KdPrint(("[yhy] DriverEntry IoCreateDevice ErrorCode:%08x\n", status));
        return status;
    }

    // 使用缓冲区IO方式
    //pDevice->Flags |= DO_BUFFERED_IO;
    pDevice->Flags |= DO_DIRECT_IO;

    // 3. 创建符号
    UNICODE_STRING symbolName;
    RtlInitUnicodeString(&symbolName, SYMBOLIC_LINK);
    status = IoCreateSymbolicLink(&symbolName, &deviceName);
    if (!NT_SUCCESS(status)) {
        KdPrint(("[yhy] DriverEntry IoCreateSymbolicLink ErrorCode:%08x\n", status));
        return status;
    }
    KdPrint(("[yhy] DriverEntry IoCreateSymbolicLink Success! symbolName is :%wZ", &symbolName));

    return status;
}